;---------------------------------------
;
; Boyer-Moore pattern matching
; Copyright (c) 1994 by Alessandro Scotti
;
;---------------------------------------
P386
JUMPS
LOCALS

MAXPATTLEN      EQU     64      ; Max length of pattern, should be even

PUBLIC  bmSetPattern
PUBLIC  bmExactSearch

;---------------------------------------
;
; Data segment
;
SE_DATA SEGMENT USE16 PARA PUBLIC 'DATA'
        ASSUME  ds:SE_DATA

CTable  DW      256     DUP(?)          ; Character skip table
PCount  DW      ?
PMatch  DB      MAXPATTLEN DUP(?)       ; Pattern to search

SE_DATA ENDS

;---------------------------------------
;
; Code segment
;
SE_TEXT SEGMENT USE16 PARA PUBLIC 'CODE'
        ASSUME  cs:SE_TEXT, es:NOTHING

;---------------------------------------
;
; Sets the pattern for a search.
;
; Input:
;       Pattern = pointer to pattern to search
;       PattLen = length of pattern
; Output:
;       zero if successful, otherwise a negative error code
;
bmSetPattern    PROC PASCAL FAR
        ARG     Pattern:DWORD,  \
                PattLen:WORD
        USES    ds, es, di

        mov     ax, SEG SE_DATA
        mov     ds, ax
        mov     es, ax

; Check pattern length
        mov     ax, -1
        mov     cx, [PattLen]
        jcxz    @@Exit                  ; Null pattern, exit
        cmp     cx, MAXPATTLEN
        jae     @@Exit                  ; Pattern too long, exit

; Copy pattern
        les     di, [Pattern]
        mov     bx, OFFSET PMatch
@@CopyLoop:
        mov     al, es:[di]
        mov     ds:[bx], al
        inc     di
        inc     bx
        dec     cx
        jnz     @@CopyLoop

; Initialize character table to maximum skip distance
        mov     di, OFFSET CTable
        mov     ax, [PattLen]
        mov     cx, 256
        cld
        rep     stosw

; Set skip count for characters in pattern
        mov     dx, ax                  ; Save pattern length
        dec     ax                      ; Skip over last character
        mov     [PCount], ax            ; Copy pattern length
        jz      @@Done                  ; No more characters, exit
        mov     di, WORD PTR [Pattern]
        add     di, ax
        dec     di                      ; Make indexing 0-based
@@Loop:
        mov     bl, es:[di]             ; Get pattern character
        xor     bh, bh
        shl     bx, 1
        cmp     CTable[bx], dx          ; Already initialized?
        jne     @@Next                  ; Yes, skip this character
        mov     CTable[bx], ax          ; Set character skip count
@@Next:
        dec     di                      ; Next character (moving right to left)
        dec     ax                      ; End of pattern?
        jnz     @@Loop                  ; No, continue

@@Done:
        xor     ax, ax
@@Exit:
        ret
bmSetPattern    ENDP

;---------------------------------------
;
; Searches a buffer for an exact match of the current pattern.
;
; Input:
;       Buffer  = far pointer to buffer
;       BufLen  = buffer length in characters
; Output:
;       the index of the LAST character in a match, or -1 if no match found
;
bmExactSearch   PROC PASCAL FAR
        ARG     Buffer:DWORD,   \
                BufLen:WORD
        USES    ds, si, es, di

        mov     ax, SEG SE_DATA
        mov     ds, ax

; Check buffer length
        mov     ax, -1                  ; Assume an error
        mov     cx, [BufLen]            ; Get buffer length
        jcxz    @@Exit                  ; Exit if buffer is empty
        cmp     cx, [PCount]            ; Buffer too small?
        jb      @@Exit                  ; Yes, exit

; Initialize variables
        les     di, [Buffer]            ; Get pointer to buffer
        mov     si, [PCount]            ; Get length of pattern
        mov     dx, di
        mov     ax, si                  ; Save pattern length
        add     dx, cx                  ; Get max buffer offset
        add     di, si                  ; Start matching at end of pattern

; Search loop: assuming that non-taken jumps are faster (which is
; true for the 386 and 486) the following loop favours mismatches
@@Loop:
        mov     bl, es:[di]             ; Get text character
        cmp     bl, ds:PMatch[si]       ; Does character match?
        je      @@Hit                   ; Yes, update counters
        xor     bh, bh
        shl     bx, 1
        mov     si, ax                  ; Restart at end of pattern
        add     di, CTable[bx]          ; Bump text offset
        cmp     di, dx                  ; End of buffer?
        jb      @@Loop                  ; No, continue
        mov     ax, -1                  ; Not found
        jmp     @@Exit
@@Hit:
        dec     di
        dec     si                      ; End of pattern?
        jns     @@Loop                  ; No, keep matching
        add     ax, di                  ; Get offset of match
        sub     ax, WORD PTR [Buffer]
        inc     ax

@@Exit:
        cwd                             ; Convert return code to double word
        ret
bmExactSearch   ENDP

SE_TEXT ENDS
END
